#
# Copyright 2020 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import structlog
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue

from jaeger_client import Config
from jaeger_client.constants import SAMPLER_TYPE_CONST

from pyvoltha_min.adapters.common.kvstore.twisted_etcd_store import TwistedEtcdStore
from pyvoltha_min.common.structlog_setup import string_to_int
from pyvoltha_min.adapters.log_features import GlobalTracingSupport

GLOBAL_CONFIG_ROOT_NODE = "global"
DEFAULT_KV_STORE_CONFIG_PATH = "config"
KV_STORE_DATA_PATH_PREFIX = "service/voltha"
KV_STORE_PATH_SEPARATOR = "/"
CONFIG_TYPE = "loglevel"
FEATURE_TYPE = "logfeatures"
TRACING_TYPE = "trace_publish"
CORRELATION_TYPE = "log_correlation"
FEATURE_ENABLED = b'ENABLED'
FEATURE_DISABLED = b'DISABLED'
DEFAULT_PACKAGE_NAME = "default"
DEFAULT_SUFFIX = KV_STORE_PATH_SEPARATOR + DEFAULT_PACKAGE_NAME
GLOBAL_DEFAULT_LOGLEVEL = b'WARN'
LOG_LEVEL_INT_NOT_FOUND = 0


class LogController:
    instance_id = None
    active_log_level = None

    def __init__(self, etcd_host, etcd_port, scope_manager, config, watch_callback=None):
        self.log = structlog.get_logger()
        self._initialized = False
        self._etcd_host = etcd_host
        self._etcd_port = etcd_port
        self._etcd_client = TwistedEtcdStore(self._etcd_host, self._etcd_port, KV_STORE_DATA_PATH_PREFIX)
        self.global_config_path = self.make_component_path(GLOBAL_CONFIG_ROOT_NODE) + DEFAULT_SUFFIX

        self._component_name = config.get('component_name')
        self._component_config_path = self.make_component_path(self._component_name)
        self._feature_root = self.make_root_path(self._component_name)
        self._global_log_level = GLOBAL_DEFAULT_LOGLEVEL
        self._watch_callback = watch_callback

        # Reference
        GlobalTracingSupport.set_log_controller(self, scope_manager, config)

    @staticmethod
    def make_component_path(component):
        return DEFAULT_KV_STORE_CONFIG_PATH + KV_STORE_PATH_SEPARATOR + component + \
               KV_STORE_PATH_SEPARATOR + CONFIG_TYPE

    @staticmethod
    def make_root_path(component):
        return DEFAULT_KV_STORE_CONFIG_PATH + KV_STORE_PATH_SEPARATOR + component + \
               KV_STORE_PATH_SEPARATOR + FEATURE_TYPE

    def make_config_path(self, key):
        return self._component_config_path + KV_STORE_PATH_SEPARATOR + key

    def make_feature_path(self, feature):
        return self._feature_root + KV_STORE_PATH_SEPARATOR + feature

    def is_global_config_path(self, key):
        if isinstance(key, bytes):
            key = key.decode('utf-8')

        return key is not None and \
               key[:len(KV_STORE_DATA_PATH_PREFIX)] == KV_STORE_DATA_PATH_PREFIX and \
               key[len(KV_STORE_DATA_PATH_PREFIX)+1:] == self.global_config_path

    def is_component_default_path(self, key):
        if isinstance(key, bytes):
            key = key.decode('utf-8')

        default_path = self.make_config_path(DEFAULT_PACKAGE_NAME)
        return key is not None and \
               key[:len(KV_STORE_DATA_PATH_PREFIX)] == KV_STORE_DATA_PATH_PREFIX and \
               key[len(KV_STORE_DATA_PATH_PREFIX)+1:] == default_path

    def is_feature_path(self, key):
        if isinstance(key, bytes):
            key = key.decode('utf-8')

        if self.make_feature_path(TRACING_TYPE) in key:
            return True, TRACING_TYPE

        if self.make_feature_path(CORRELATION_TYPE) in key:
            return True, CORRELATION_TYPE

        return False, None

    @inlineCallbacks
    def get_global_loglevel(self):
        loglevel = self._global_log_level
        try:
            level = yield self._etcd_client.get(self.global_config_path)
            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))

                if level_int == 0:
                    self.log.warn("unsupported-log-level", requested_level=level)
                else:
                    loglevel = level

        except KeyError:
            self.log.warn("failed-to-retrieve-log-level", path=self.global_config_path)

        returnValue(loglevel)

    @inlineCallbacks
    def get_log_features(self):
        features = dict()
        try:
            features_info = yield self._etcd_client.get(self.feature_path)
            if features_info is not None:
                # level_int = string_to_int(str(level, 'utf-8'))
                #
                # if level_int == 0:
                #     self.log.warn("unsupported-log-level", requested_level=level)
                # else:
                #     loglevel = level
                pass
            # TODO: Convert to dictionary
        except KeyError:
            self.log.warn("failed-to-retrieve-log-level", path=self.global_config_path)

        returnValue(features)

    @inlineCallbacks
    def get_component_default_loglevel(self, global_default_loglevel):
        self.log.info("entry", path=self._component_config_path)
        component_default_loglevel = global_default_loglevel

        try:
            path = self.make_config_path(DEFAULT_PACKAGE_NAME)
            level = yield self._etcd_client.get(path)

            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))

                if level_int == 0:
                    self.log.warn("unsupported-log-level", unsupported_level=level)

                else:
                    component_default_loglevel = level
                    self.log.info("default-loglevel", new_level=level)

        except KeyError:
            self.log.warn("failed-to-retrieve-log-level",
                          path=self._component_config_path + DEFAULT_SUFFIX)

        if component_default_loglevel == "":
            component_default_loglevel = GLOBAL_DEFAULT_LOGLEVEL.encode('utf-8')

        returnValue(component_default_loglevel)

    @inlineCallbacks
    def get_package_loglevel_int(self, package):
        self.log.debug("entry", path=self._component_config_path)
        level_int = LOG_LEVEL_INT_NOT_FOUND
        path = self.make_config_path(package)
        try:
            level = yield self._etcd_client.get(path)
            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))
                if level_int == 0:
                    self.log.warn("unsupported-log-level", unsupported_level=level)
                    level_int = LOG_LEVEL_INT_NOT_FOUND

        except KeyError:
            self.log.warn("failed-to-retrieve-log-level", path=path)

        returnValue(level_int)

    @inlineCallbacks
    def start_watch_log_config_change(self, instance_id, initial_default_loglevel):
        self.log.debug("entry")
        LogController.instance_id = instance_id

        if self._component_name is None:
            raise Exception("Invalid pod component name")

        _results = yield self._etcd_client.watch(self.global_config_path, self.watch_callback)
        _results = yield self._etcd_client.watch(self._component_config_path, self.watch_callback,
                                                 watch_prefix=True)
        _results = yield self._etcd_client.watch(self._feature_root, self.watch_callback,
                                                 watch_prefix=True)
        _results = yield self.set_default_loglevel(self.global_config_path,
                                                   self._component_config_path,
                                                   initial_default_loglevel.upper())

    def watch_callback(self, event):
        reactor.callFromThread(self.process_log_config_change, event)

    @inlineCallbacks
    def process_log_config_change(self, watch_event=None):
        # Forward to user callback
        if self._watch_callback is not None:
            results = yield self._watch_callback(watch_event)
            returnValue(results)

        returnValue(False)

    @inlineCallbacks
    def set_default_loglevel(self, global_config_path, component_config_path, initial_default_loglevel):
        self.log.debug('set-default-level', global_path=global_config_path,
                       component_path=component_config_path,
                       def_level=initial_default_loglevel)

        if (yield self._etcd_client.get(global_config_path)) is None:
            _results = yield self._etcd_client.set(global_config_path, GLOBAL_DEFAULT_LOGLEVEL)

        default_path = self.make_config_path(DEFAULT_PACKAGE_NAME)
        if (yield self._etcd_client.get(default_path)) is None:
            _results = yield self._etcd_client.set(default_path, initial_default_loglevel)

    @inlineCallbacks
    def persist_initial_log_features(self):
        log_correlation = FEATURE_ENABLED if GlobalTracingSupport.log_correlation_status else FEATURE_DISABLED
        trace_enabled = FEATURE_ENABLED if GlobalTracingSupport.trace_publishing_status else FEATURE_DISABLED

        self.log.debug('presist-log-features', feature_root=self._feature_root,
                       correlation=log_correlation, tracing=trace_enabled)
        try:
            path = self.make_feature_path(CORRELATION_TYPE)
            status = yield self._etcd_client.get(path)
            if status != log_correlation:
                _results = yield self._etcd_client.set(path, log_correlation)

            path = self.make_feature_path(TRACING_TYPE)
            status = yield self._etcd_client.get(path)
            if status != trace_enabled:
                _results = yield self._etcd_client.set(path, trace_enabled)

        except Exception as e:
            self.log.exception('initial-persist-failed', e=e)
            # Return none always

    def enable_jaeger_client(self):
        log_correlation_enabled = GlobalTracingSupport.log_correlation_status
        tracing_enabled = GlobalTracingSupport.trace_publishing_status

        if tracing_enabled or log_correlation_enabled:
            # The following are available 'config' dictionary options
            #
            # logging
            #   - If true, the jaeger tracing structured logger (jaeger_tracing component) is also provided
            #     to the ErrorReporter for any RemoteThrottler and a CompositeReporter is provided to the
            #     Tracer
            #   - Default: False
            #
            # local_agent (dict)
            #   - Default:
            #      - 'enabled' (bool), Default: LOCAL_AGENT_DEFAULT_ENABLED = True
            #      - 'sampling_port' , Default: DEFAULT_SAMPLING_PORT = 5778
            #      - 'reporting_port', Default: DEFAULT_REPORTING_PORT = 6831
            #      - 'reporting_host', Default: DEFAULT_REPORTING_HOST = 'localhost' if environment variable
            #                                   'JAEGER_AGENT_HOST' is not defined
            #
            # sampler
            #   - Sampler config (dict) A sampler is responsible for deciding if a particular
            #                           span should be "sampled", i.e. recorded in permanent storage. Besides
            #                           the ones that can be defined via config, Jaeger also has several
            #                           additional Samplers available including AdaptiveSampler,
            #                           RemoteControlledSampler, GuaranteedThroughputProbabilisticSampler
            #     - 'type': Sampler type (str) See below
            #        - default: None
            #             - SAMPLER_TYPE_CONST = "const" A sampler always returns the same decision.
            #                  - param -> decision (bool)
            #                             default: False
            #
            #             - SAMPLER_TYPE_PROBABILISTIC = "PROBABILISTIC". A sampler that randomly samples
            #                             a certain percentage of traces specified by the samplingRate, in
            #                             the range between 0.0 and 1.0.
            #                              It relies on the fact that new trace IDs are 64bit random numbers
            #                              themselves, thus making the sampling decision without generating
            #                              a new random number, but simply calculating if
            #                              traceID < (samplingRate * 2^64). Note that we actually
            #                              (zero out) the most significant bit.
            #                  - param -> rate (float) Sampling rate must be between 0.0 and 1.0
            #                             default: False
            #
            #             - SAMPLER_TYPE_RATE_LIMITING = "RATE_LIMITING" or 'rate_limiting' Samples at most
            #                              max_traces_per_second. The distribution of sampled traces follows
            #                              burstiness of the service, i.e. a service with uniformly distributed
            #                              requests will have those requests sampled uniformly as well, but if
            #                              requests are bursty, especially sub-second, then a number of
            #                              sequential requests can be sampled each second.
            #                  - param -> max_traces_per_second (float)
            #                             default: 10
            #
            #   - 'param': Parameter to provide to sampler type. See defined sampler types above
            #              for more information.
            #              - default: None
            # tags
            #   - Process-wide tracer tags
            #   - Default: None
            #
            # enabled
            #   - (bool) Defined, but I do not see it in use any where in Jaeger-cient
            #   - Default: True
            #
            # reporter_batch_size
            #   - How many spans a remote reporter can submit at once to Collector
            #   - Default: 10
            #
            # reporter_queue_size
            #   - Remote reporter queue size.  How many spans a remote reporter can hold
            #     in memory before starting to drop spans
            #   - Default: 100
            #
            # propagation
            #   - (str) if equal to 'b3' replace the codec with a B3 enabled instance
            #   - Default: None
            #
            # max_tag_value_length
            #   - Max allowed tag value length. Longer values will be truncated.
            #   - Default: MAX_TAG_VALUE_LENGTH = 1024
            #
            # max_traceback_length
            #   -  Max length for traceback data. Longer values will be truncated.
            #   - Default: MAX_TRACEBACK_LENGTH = 4096
            #
            # reporter_flush_interval
            #   - How often remote reporter does a preemptive flush of its buffers, A remote reporter
            #     receives completed spans from Tracer and submits them out of process
            #   - Default: DEFAULT_FLUSH_INTERVAL = 1
            #
            # sampling_refresh_interval
            #   - interval in seconds for polling for strategy
            #   - Default: DEFAULT_SAMPLING_INTERVAL = 5
            #
            # trace_id_header
            #   - The name of the HTTP header used to encode trace ID
            #   - Default: TRACE_ID_HEADER -> 'uber-trace-id' if six.PY3 else b'uber-trace-id'
            #
            # generate_128bit_trace_id
            #   - Generate 128 bit trace id, otherwise 16
            #   - Default: False
            #   - Override: True if env variable 'JAEGER_TRACEID_128BIT' == 'true'
            #
            # baggage_header_prefix
            #   - The prefix for HTTP headers used to record baggage
            #   - Default: 'uberctx-' if six.PY3 else b'uberctx-'
            #
            # service_name
            #   - Service Name. Used to override service name defined in 'service_name' parameter to
            #                   the Config() initializer
            #   - Default: None
            #
            # throttler
            #   - Throttler Group (dict). If not specified, a default of 'None' is used which specifies
            #                             that a remote throttler will be created.  If created a Remote
            #                             Throttler controls the flow of spans emitted from client to
            #                             prevent flooding. RemoteThrottler requests credits from the
            #                             throttling service periodically. These credits determine the
            #                             amount of debug spans a client may emit for a particular
            #                             operation without receiving more credits.
            #      - 'port':             (int) DEFAULT_THROTTLER_PORT = DEFAULT_SAMPLING_PORT = 5778
            #      - 'refresh_interval': (int) DEFAULT_THROTTLER_INTERVAL = 5
            #   - Default: None
            #
            address = GlobalTracingSupport.active_trace_agent_address.split(':')

            config = Config(
                config={  # usually read from some yaml config
                    'sampler': {
                        'type': SAMPLER_TYPE_CONST,
                        'param': tracing_enabled,
                    },
                    'local_agent': {
                        'reporting_host': address[0],
                        'reporting_port': int(address[1]),
                    },
                    'logging': True,
                },
                service_name=GlobalTracingSupport.component_name,
                scope_manager=GlobalTracingSupport.scope_manager,
                validate=True,
            )
            # this call also sets opentracing.tracer
            if self._initialized:
                config.new_tracer()
            else:
                self._initialized = True
                config.initialize_tracer()



# Copyright 2017-present Adtran, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from voltha_protos.events_pb2 import EventCategory, EventSubCategory

from pyvoltha_min.adapters.extensions.events.adapter_events import DeviceEventBase

"""
    OnuSignalsFailureIndication {
                fixed32 intf_id = 1;
                fixed32 onu_id = 2;
                string status = 3;
                fixed32 inverse_bit_error_rate = 4;
"""


class OnuSignalFailEvent(DeviceEventBase):
    def __init__(self, event_mgr, onu_id, intf_id, inverse_bit_error_rate, raised_ts):
        super(OnuSignalFailEvent, self).__init__(event_mgr, raised_ts, object_type='ONU SIGNALS FAIL',
                                                 event='ONU_SIGNALS_FAIL',
                                                 category=EventCategory.COMMUNICATION,
                                                 sub_category=EventSubCategory.PON)
        self._onu_id = onu_id
        self._intf_id = intf_id
        self._inverse_bit_error_rate = inverse_bit_error_rate

    def get_context_data(self):
        return {'onu-id': self._onu_id,
                'intf-id': self._intf_id,
                'inverse-bit-error-rate': self._inverse_bit_error_rate}



# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import functools
import structlog

from twisted.internet import defer
from opentracing import global_tracer, tags
from jaeger_client import Span as JaegerSpan, SpanContext as JaegerSpanContext

from pyvoltha_min.adapters.log_features import GlobalTracingSupport

ROOT_SPAN_KEY_NAME = "op-name"

log = structlog.get_logger()


def create_async_span(span_name, ignore_active_span=False, **kwargs):
    """
    Creates a Async Child Span with Follows-From relationship from Parent Span embedded
    in passed context.

    Should be used only in scenarios when
      a) There is dis-continuation in execution and thus result of Child span does
          not affect the Parent flow at all
      b) The execution of Child Span is guaranteed to start after the completion
         of Parent Span

    In case of any confusion, use CreateChildSpan method

    Some situations where this method would be suitable includes Kafka Async RPC
    call, Propagation of Event across a channel etc.
    """
    tracer = global_tracer()

    if not (GlobalTracingSupport.log_correlation_status or GlobalTracingSupport.trace_publishing_status):
        return tracer.start_span(span_name, ignore_active_span=ignore_active_span)

    parent_span = tracer.active_span
    child_span = tracer.start_span(span_name, references=parent_span, ignore_active_span=ignore_active_span)

    if parent_span is None or not parent_span.get_baggage_item(ROOT_SPAN_KEY_NAME):
        child_span.set_baggage_item(ROOT_SPAN_KEY_NAME, span_name)

    enrich_span(child_span, **kwargs)
    return child_span


def create_child_span(task_name, ignore_active_span=False, **kwargs):
    """
    Creates a Child Span from Parent Span embedded in current context context. Should be used before
    starting a new major operation in Synchronous or Asynchronous mode (go routine), such as following:

        1. Start of all implemented External API methods unless using a interceptor for auto-injection
           of Span (Server side impl)
        2. Just before calling an Third-Party lib which is invoking a External API (etcd, kafka)
        3. In start of a Go Routine responsible for performing a major task involving significant duration
        4. Any method which is suspected to be time consuming...
    """
    tracer = global_tracer()

    parent_span = tracer.active_span
    child_span = tracer.start_span(task_name, child_of=parent_span, ignore_active_span=ignore_active_span)
    #     parentSpan := opentracing.SpanFromContext(ctx)
    #     childSpan, newCtx := opentracing.StartSpanFromContext(ctx, taskName)

    if parent_span is None or not parent_span.get_baggage_item(ROOT_SPAN_KEY_NAME):
        child_span.set_baggage_item(ROOT_SPAN_KEY_NAME, task_name)

    enrich_span(child_span, **kwargs)
    return child_span


def enrich_span(span, **kwargs):
    """ Method to inject additional log fields into Span e.g. device-id """

    if isinstance(span, JaegerSpan):
        # Inject as a BaggageItem when the Span is the Root Span so that it propagates
        # across the components along with Root Span (called as Trace)
        # else, inject as a Tag so that it is attached to the Child Task
        is_root_span = span.trace_id == span.span_id

        for _, field in kwargs.items():
            for key, value in field.items():
                if is_root_span:
                    span.set_baggage_item(key, str(value))
                else:
                    span.set_tag(key, value)


def span_from_context(ctx):
    """ Twisted python equivalent of golang opentracing.SpanFromContext(ctx) """

    if not isinstance(ctx, JaegerSpanContext):
        return None

    # Currently only support if in main thread.
    if not threading.main_thread():
        return None

    # TODO: Do rest of it here
    return None


def extract_context_attributes():
    """
    Extracts details of current context as log fields from the Tracing Span
    injected into the context instance. Following log fields are extracted:

        -  Operation Name : key as 'op-name' and value as Span operation name
        -  Operation Id : key as 'op-id' and value as 64 bit Span Id in hex digits string

    Additionally, any tags present in Span are also extracted to use as log
    fields e.g. device-id.

    If no Span is found associated with context, blank slice is returned without
    any log fields

    :return: (dict) of fields or None
    """
    if not GlobalTracingSupport.log_correlation_status:
        return None

    span = global_tracer().active_span
    if not isinstance(span, JaegerSpan):
        return None

    op_id = span.trace_id & 0xFFFFFFFFFFFFFFFF   # Lower 64-bits only
    op_name = span.get_baggage_item(ROOT_SPAN_KEY_NAME)
    task_id = span.span_id
    task_name = span.operation_name

    if not op_name:
        span.set_baggage_item(ROOT_SPAN_KEY_NAME, task_name)
        op_name = task_name

    fields = {'op-id': '{:016x}'.format(op_id), 'op-name': op_name}

    if task_id != op_id:
        fields['task-id'] = '{:016x}'.format(task_id)
        fields['task-name'] = task_name

    try:
        for key, value in span.tags:
            # Ignore the special tags added by Jaeger, middleware (sampler.type, span.*)
            # present in the span
            if any(key.startswith(prefix) for prefix in ('sampler.', 'span.', 'component.')):
                continue

            fields[key] = value

    except Exception as _e:
        pass

    try:
        for key, value in span.context.baggage.items():
            if key != 'rpc-span-name':
                fields[key] = value

    except Exception as _e:
        pass

    return fields


def mark_span_error(error=None):
    """ Inject error information into the currently active span """

    span = global_tracer().active_span
    if span is not None:
        span.set_tag(tags.ERROR, True)

        if error is not None:
            span.set_tag('err', error)


def traced_function(func=None, name=None, on_start=None, ignore_active_span=False,
                    require_active_trace=False, async_result=False):
    """
    A decorator that enables tracing of the wrapped function or
    twisted deferred routine provided there is a parent span already established.
    .. code-block:: python
        @traced_function
        def my_function1(arg1, arg2=None)
            ...
    :param func: decorated function or Twisted deferred routine

    :param name: optional name to use as the Span.operation_name.
                 If not provided, func.__name__ will be used.

    :param on_start: an optional callback to be executed once the child span
                     is started, but before the decorated function is called. It can be
                     used to set any additional tags on the span, perhaps by inspecting
                     the decorated function arguments. The callback must have a signature
                     `(span, *args, *kwargs)`, where the last two collections are the
                     arguments passed to the actual decorated function.
                     .. code-block:: python
                         def extract_call_site_tag(span, *args, *kwargs)
                             if 'call_site_tag' in kwargs:
                                 span.set_tag('call_site_tag', kwargs['call_site_tag'])
                         @traced_function(on_start=extract_call_site_tag)
                         @inlineCallback             TODO: InlineCallback not yet supported fully/tested
                         def my_function(arg1, arg2=None, call_site_tag=None)
                ...
    :param require_active_trace: controls what to do when there is no active
                                 trace. If require_active_trace=True, then no span is created.
                                 If require_active_trace=False, a new trace is started.

    :param async_result: if a parent trace is present, this specifies that the
                         span will (may) finish asynchronously (possibly after) the parent span
                         and should be treated as more as 'follows-from'

    :param ignore_active_span: if true, start a new root span, ignoring any currently active one

    :return: returns a tracing decorator
    """
    if func is None:
        return functools.partial(traced_function, name=name, on_start=on_start,
                                 ignore_active_span=ignore_active_span,
                                 require_active_trace=require_active_trace,
                                 async_result=async_result)

    operation_name = name if name else func.__name__

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        parent_span = global_tracer().active_span if not ignore_active_span else None
        if parent_span is None and require_active_trace:
            return func(*args, **kwargs)

        tracer = global_tracer()
        if async_result:
            reference_span, parent_span = parent_span, None
        else:
            reference_span = None

        with tracer.start_active_span(operation_name, child_of=parent_span,
                                      references=reference_span, ignore_active_span=ignore_active_span,
                                      finish_on_close=False) as scope:
            span = scope.span

            if callable(on_start):
                on_start(span, *args, **kwargs)

            # We explicitly invoke deactivation callback for the StackContext,
            # because there are scenarios when it gets retained forever, for
            # example when a Periodic Callback is scheduled lazily while in the
            # scope of a tracing StackContext.
            try:
                d = func(*args, **kwargs)

                # Twisted routines and inlineCallbacks generate defers
                if isinstance(d, defer.Deferred):
                    def done(results):
                        span.finish()
                        return results

                    def failed(reason):
                        # Set error in span
                        span.log(event='exception', payload=reason)
                        span.set_tag('error', 'true')
                        span.finish()
                        return reason

                    if d.called:
                        span.finish()
                    else:
                        d.addCallbacks(done, failed)
                else:
                    span.finish()
                return d

            except Exception as e:
                span.set_tag(tags.ERROR, 'true')
                span.log(event='exception', payload=e)
                span.finish()
                raise

    return decorator



# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import codecs

import etcd3
import structlog
from twisted.internet.defer import inlineCallbacks

from pyvoltha_min.common.utils.asleep import asleep


class EtcdStore:
    """ Config kv store for etcd with a cache for quicker subsequent reads

        TODO: This will block the reactor. Should either change
        whole call stack to yield or put the put/delete transactions into a
        queue to write later with twisted. Will need a transaction
        log to ensure we don't lose anything.
        Making the whole callstack yield is troublesome because other tasks can
        come in on the side and start modifying things which could be bad.
    """

    RETRY_BACKOFF = [0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    def __init__(self, host, port, path_prefix):

        self.log = structlog.get_logger()
        self._etcd = etcd3.client(host=host, port=port)
        self.host = host
        self.port = port
        self._path_prefix = path_prefix
        self.retries = 0

    def make_path(self, key):
        return '{}/{}'.format(self._path_prefix, key)

    def __getitem__(self, key):
        (value, _meta) = self._kv_get(self.make_path(key))
        if value is not None:
            return value
        raise KeyError(key)

    def __contains__(self, key):
        (value, _meta) = self._kv_get(self.make_path(key))
        if value is not None:
            return True
        return False

    def __setitem__(self, key, value):
        try:
            self._kv_put(self.make_path(key), value)
        except Exception as e:
            self.log.exception('cannot-set-item', e=e)

    def __delitem__(self, key):
        self._kv_delete(self.make_path(key))

    def list(self, key):
        return self._kv_list(self.make_path(key))

    @inlineCallbacks
    def _backoff(self, msg):
        wait_time = self.RETRY_BACKOFF[min(self.retries,
                                           len(self.RETRY_BACKOFF) - 1)]
        self.retries += 1
        self.log.error(msg, retry_in=wait_time)
        yield asleep(wait_time)

    def _redo_etcd_connection(self):
        self._etcd = etcd3.client(host=self.host, port=self.port)

    def _clear_backoff(self):
        if self.retries:
            self.log.info('reconnected-to-etcd', after_retries=self.retries)
            self.retries = 0

    def _get_etcd(self):
        return self._etcd

    # Proxy methods for etcd with retry support
    def _kv_get(self, *args, **kw):
        return self._retry('GET', *args, **kw)

    def _kv_put(self, *args, **kw):
        return self._retry('PUT', *args, **kw)

    def _kv_delete(self, *args, **kw):
        return self._retry('DELETE', *args, **kw)

    def _kv_list(self, *args, **kw):
        return self._retry('LIST', *args, **kw)

    def _retry(self, operation, *args, **kw):

        # etcd data sometimes contains non-utf8 sequences, replace
        self.log.debug('backend-op',
                       operation=operation,
                       args=[codecs.encode(x, 'utf8', 'replace') for x in args],
                       kw=kw)

        while 1:
            try:
                etcd = self._get_etcd()
                self.log.debug('etcd', etcd=etcd, operation=operation,
                               args=[codecs.encode(x, 'utf8', 'replace') for x in args])
                if operation == 'GET':
                    (value, meta) = etcd.get(*args, **kw)
                    result = (value, meta)
                elif operation == 'PUT':
                    result = etcd.put(*args, **kw)
                elif operation == 'DELETE':
                    result = etcd.delete(*args, **kw)
                elif operation == 'LIST':
                    result = etcd.get_prefix(*args, *kw)
                else:
                    # Default case - consider operation as a function call
                    result = operation(*args, **kw)
                self._clear_backoff()
                break
            except Exception as e:
                self.log.exception(e)
                self._backoff('unknown-error-with-etcd')
            self._redo_etcd_connection()

        return result



#
# Copyright 2020 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import structlog
from opentracing import global_tracer

log = structlog.get_logger()


class _LogFeatures:
    def __init__(self):
        self._log_correlation_enabled = None
        self._trace_publishing_enabled = None
        self._component_name = None
        self._active_trace_agent_address = None
        self._scope_manager = None
        self._log_controller = None

    @property
    def log_correlation_status(self):
        return self._log_correlation_enabled is not None and self._log_correlation_enabled

    @log_correlation_status.setter
    def log_correlation_status(self, value):
        if not isinstance(value, bool):
            raise TypeError('Boolean required')

        if value != self._log_correlation_enabled:
            # Close any previous tracer and reopen if required
            tracer = global_tracer()
            if hasattr(tracer, 'close'):
                tracer.close()

            self._log_correlation_enabled = value
            self._log_controller.enable_jaeger_client()

    @property
    def trace_publishing_status(self):
        return self._trace_publishing_enabled is not None and self._trace_publishing_enabled

    @trace_publishing_status.setter
    def trace_publishing_status(self, value):
        if not isinstance(value, bool):
            raise TypeError('Boolean required')

        if value != self._trace_publishing_enabled:
            # Close any previous tracer and reopen if required
            tracer = global_tracer()
            if hasattr(tracer, 'close'):
                tracer.close()

            self._trace_publishing_enabled = value
            self._log_controller.enable_jaeger_client()

    @property
    def component_name(self):
        return self._component_name

    @property
    def active_trace_agent_address(self):
        return self._active_trace_agent_address

    @property
    def scope_manager(self):
        return self._scope_manager

    @scope_manager.setter
    def scope_manager(self, value):
        if self._scope_manager is None:
            self._scope_manager = value

    def set_log_controller(self, log_controller, scope_manager, config):
        self._log_controller = log_controller
        self._active_trace_agent_address = config.get('trace_agent_address')
        self._component_name = config.get('component_name', os.environ.get('COMPONENT_NAME',
                                                                           'unknown-olt'))
        self.scope_manager = scope_manager
        # Set trace/log correlation directly.  During initial startup, we want to start the
        # OpenTracing client at the end of our init and not from the property setter functions.
        self._trace_publishing_enabled = config.get('trace_enabled')
        self._log_correlation_enabled = config.get('log_correlation_enabled')


GlobalTracingSupport = _LogFeatures()
